#include "Man.h"
#include "Map.h"
#include "ScanKeys.h"
#include "Chars.h"
#include "Vram.h"
#include "VVram.h"
#include "Main.h"
#include "Knife.h"
#include "Sound.h"

struct Direction {
    byte key;
    sbyte dx, dy;
    byte pattern;
    byte index;
    sbyte xOffset, yOffset;
};
static const Direction[] Directions = {
    { Keys_Left, -1, 0, Pattern_Man_Left, 0, 0, 0 },
    { Keys_Right, 1, 0, Pattern_Man_Right, 1, 1, 0 },
    { Keys_Up, 0, -1, Pattern_Man_Up, 2, 0, 0 },
    { Keys_Down, 0, 1, Pattern_Man_Down, 3, 0, 1 },
};

byte ManX, ManY;
bool ManLost;
static ptr<Direction> pPrevDirection;
static byte pattern;
static byte seqIndex;
static bool buttonOn;

static void UpdateBasePosition()
{
    if (ManX < VVramWidth / 2) {
        BaseX = 0;
    }
    else {
        byte x, margin;
        margin = MapWidth - VVramWidth;
        x = ManX - VVramWidth / 2;
        if (x >= margin) {
            BaseX = margin;
        }
        else {
            BaseX = x;
        }
    }
    if (ManY < VVramHeight / 2) {
        BaseY = 0;
    }
    else {
        byte y, margin;
        margin = MapHeight - VVramHeight;
        y = ManY - VVramHeight / 2;
        if (y >= margin) {
            BaseY = margin;
        }
        else {
            BaseY = y;
        }
    }
}

void InitMan()
{
    ManX = ToMapSize(*pStage); ++pStage;
    ManY = ToMapSize(*pStage); ++pStage;
    UpdateBasePosition();
    pPrevDirection = Directions + 1; // right
    pattern = pPrevDirection->pattern;
    seqIndex = 0;
}


void DrawMan()
{
    CopyPattern(Char_Man, pattern, 4);
    VPut2(ManX, ManY, Char_Man);
}


void MoveMan()
{
    byte key;
    ptr<Direction> pDirection;
    byte x, y;
    if (ManLost) return;

    key = ScanKeys();
    for (pDirection : Directions) {
        if ((key & pDirection->key) != 0) {
            x = ManX + pDirection->dx;
            y = ManY + pDirection->dy;
            if (CanMove2(x, y)) {
                pPrevDirection = pDirection;
                goto canMove;
            }
        }
    }
    if ((key & Keys_Dir) != 0) {
        x = ManX + pPrevDirection->dx;
        y = ManY + pPrevDirection->dy;
        if (CanMove2(x, y)) {
            canMove:
            ManX = x;
            ManY = y;
            UpdateBasePosition();
            if (x == GoalX && y == GoalY) {
                StageClear = true;
            }
            seqIndex = (seqIndex + 1) & 3;
            if (seqIndex == 3) {
                seqIndex = 1;
            }
            pattern = pPrevDirection->pattern + ((1 + seqIndex) << 2);
            goto testButton;
        }
    }
    pattern = pPrevDirection->pattern;
    testButton:
    if ((key & Keys_Button0) != 0 && !buttonOn) {
        byte x, y;
        x = ManX + pPrevDirection->dx;
        y = ManY + pPrevDirection->dy;
        if (CanMove1(x, y)) {
            StartKnife(x, y, pPrevDirection->index);
            buttonOn = true;
            return;
        }
    }
    else {
        buttonOn = false;
    }
}


void EndMan()
{
    static const byte[] Patterns = {
        Pattern_Man_Left_Stop,
        Pattern_Man_Down_Stop,
        Pattern_Man_Die2,
        Pattern_Man_Die3
    };
    byte i;
    repeat (10) {
        CopyPattern(Char_Man, Patterns[i], 4);
        VPut2(ManX, ManY, Char_Man);
        VVramToVram();
        Sound_Loose();
        i = (i + 1) & 3;
    }
}